var _a;
import { RawAccount } from "cojson";
import { Account, Group, InitValues, ItemsSym, Ref, SchemaInit, co, ensureCoValueLoaded, inspect, isRefEncoded, loadCoValue, loadCoValueEf, makeRefs, subscribeToCoValue, subscribeToCoValueEf, subscribeToExistingCoValue, } from "../internal.js";
import { encodeSync, decodeSync } from "@effect/schema/Schema";
/**
 * CoLists are collaborative versions of plain arrays.
 *
 *  * @categoryDescription Content
 * You can access items on a `CoList` as if they were normal items on a plain array, using `[]` notation, etc.
 *
 * Since `CoList` is a subclass of `Array`, you can use all the normal array methods like `push`, `pop`, `splice`, etc.
 *
 * ```ts
 * colorList[0];
 * colorList[3] = "yellow";
 * colorList.push("Kawazaki Green");
 * colorList.splice(1, 1);
 * ```
 *
 * @category CoValues
 */
// eslint-disable-next-line @typescript-eslint/no-explicit-any
export class CoList extends Array {
    /**
     * Declare a `CoList` by subclassing `CoList.Of(...)` and passing the item schema using `co`.
     *
     * @example
     * ```ts
     * class ColorList extends CoList.Of(
     *   co.string
     * ) {}
     * class AnimalList extends CoList.Of(
     *   co.ref(Animal)
     * ) {}
     * ```
     *
     * @category Declaration
     */
    static Of(item) {
        var _b, _c;
        // TODO: cache superclass for item class
        return _c = class CoListOf extends _a {
                constructor() {
                    super(...arguments);
                    this[_b] = item;
                }
            },
            _b = co.items,
            _c;
    }
    /**
     * @ignore
     * @deprecated Use UPPERCASE `CoList.Of` instead! */
    static of(..._args) {
        throw new Error("Can't use Array.of with CoLists");
    }
    /** @internal */
    get _schema() {
        return this.constructor._schema;
    }
    /** @category Collaboration */
    get _owner() {
        return this._raw.group instanceof RawAccount
            ? Account.fromRaw(this._raw.group)
            : Group.fromRaw(this._raw.group);
    }
    /**
     * If a `CoList`'s items are a `co.ref(...)`, you can use `coList._refs[i]` to access
     * the `Ref` instead of the potentially loaded/null value.
     *
     * This allows you to always get the ID or load the value manually.
     *
     * @example
     * ```ts
     * animals._refs[0].id; // => ID<Animal>
     * animals._refs[0].value;
     * // => Animal | null
     * const animal = await animals._refs[0].load();
     * ```
     *
     * @category Content
     **/
    get _refs() {
        return makeRefs((idx) => this._raw.get(idx), () => Array.from({ length: this._raw.entries().length }, (_, idx) => idx), this._loadedAs, (_idx) => this._schema[ItemsSym]);
    }
    get _edits() {
        throw new Error("Not implemented");
    }
    get _loadedAs() {
        return Account.fromNode(this._raw.core.node);
    }
    static get [Symbol.species]() {
        return Array;
    }
    constructor(options) {
        super();
        Object.defineProperty(this, "_instanceID", {
            value: `instance-${Math.random().toString(36).slice(2)}`,
            enumerable: false,
        });
        if ("owner" in options) {
            this[InitValues] = {
                init: options.init,
                owner: options.owner,
            };
        }
        else if ("fromRaw" in options) {
            Object.defineProperties(this, {
                id: {
                    value: options.fromRaw.id,
                    enumerable: false,
                },
                _raw: { value: options.fromRaw, enumerable: false },
            });
        }
        return new Proxy(this, CoListProxyHandler);
    }
    /**
     * Create a new CoList with the given initial values and owner.
     *
     * The owner (a Group or Account) determines access rights to the CoMap.
     *
     * The CoList will immediately be persisted and synced to connected peers.
     *
     * @example
     * ```ts
     * const colours = ColorList.create(
     *   ["red", "green", "blue"],
     *   { owner: me }
     * );
     * const animals = AnimalList.create(
     *   [cat, dog, fish],
     *   { owner: me }
     * );
     * ```
     *
     * @category Creation
     **/
    static create(items, options) {
        return new this({ init: items, owner: options.owner });
    }
    push(...items) {
        for (const item of toRawItems(items, this._schema[ItemsSym])) {
            this._raw.append(item);
        }
        return this._raw.entries().length;
    }
    unshift(...items) {
        for (const item of toRawItems(items, this._schema[ItemsSym])) {
            this._raw.prepend(item);
        }
        return this._raw.entries().length;
    }
    pop() {
        const last = this[this.length - 1];
        this._raw.delete(this.length - 1);
        return last;
    }
    shift() {
        const first = this[0];
        this._raw.delete(0);
        return first;
    }
    splice(start, deleteCount, ...items) {
        const deleted = this.slice(start, start + deleteCount);
        for (let idxToDelete = start + deleteCount - 1; idxToDelete >= start; idxToDelete--) {
            this._raw.delete(idxToDelete);
        }
        let appendAfter = Math.max(start - 1, 0);
        for (const item of toRawItems(items, this._schema[ItemsSym])) {
            console.log(this._raw.asArray(), appendAfter);
            this._raw.append(item, appendAfter);
            appendAfter++;
        }
        return deleted;
    }
    toJSON() {
        const itemDescriptor = this._schema[ItemsSym];
        if (itemDescriptor === "json") {
            return this._raw.asArray();
        }
        else if ("encoded" in itemDescriptor) {
            return this._raw
                .asArray()
                .map((e) => encodeSync(itemDescriptor.encoded)(e));
        }
        else if (isRefEncoded(itemDescriptor)) {
            return this.map((item) => item?.toJSON());
        }
        else {
            return [];
        }
    }
    [inspect]() {
        return this.toJSON();
    }
    /** @category Internals */
    static fromRaw(raw) {
        return new this({ fromRaw: raw });
    }
    /** @internal */
    static schema(def) {
        this._schema || (this._schema = {});
        Object.assign(this._schema, def);
    }
    /**
     * Load a `CoList` with a given ID, as a given account.
     *
     * `depth` specifies if item CoValue references should be loaded as well before resolving.
     * The `DeeplyLoaded` return type guarantees that corresponding referenced CoValues are loaded to the specified depth.
     *
     * You can pass `[]` or for shallowly loading only this CoList, or `[itemDepth]` for recursively loading referenced CoValues.
     *
     * Check out the `load` methods on `CoMap`/`CoList`/`CoStream`/`Group`/`Account` to see which depth structures are valid to nest.
     *
     * @example
     * ```ts
     * const animalsWithVets =
     *   await ListOfAnimals.load(
     *     "co_zdsMhHtfG6VNKt7RqPUPvUtN2Ax",
     *     me,
     *     [{ vet: {} }]
     *   );
     * ```
     *
     * @category Subscription & Loading
     */
    static load(id, as, depth) {
        return loadCoValue(this, id, as, depth);
    }
    /**
     * Effectful version of `CoList.load()`.
     *
     * Needs to be run inside an `AccountCtx` context.
     *
     * @category Subscription & Loading
     */
    static loadEf(id, depth) {
        return loadCoValueEf(this, id, depth);
    }
    /**
     * Load and subscribe to a `CoList` with a given ID, as a given account.
     *
     * Automatically also subscribes to updates to all referenced/nested CoValues as soon as they are accessed in the listener.
     *
     * `depth` specifies if item CoValue references should be loaded as well before calling `listener` for the first time.
     * The `DeeplyLoaded` return type guarantees that corresponding referenced CoValues are loaded to the specified depth.
     *
     * You can pass `[]` or for shallowly loading only this CoList, or `[itemDepth]` for recursively loading referenced CoValues.
     *
     * Check out the `load` methods on `CoMap`/`CoList`/`CoStream`/`Group`/`Account` to see which depth structures are valid to nest.
     *
     * Returns an unsubscribe function that you should call when you no longer need updates.
     *
     * Also see the `useCoState` hook to reactively subscribe to a CoValue in a React component.
     *
     * @example
     * ```ts
     * const unsub = ListOfAnimals.subscribe(
     *   "co_zdsMhHtfG6VNKt7RqPUPvUtN2Ax",
     *   me,
     *   { vet: {} },
     *   (animalsWithVets) => console.log(animalsWithVets)
     * );
     * ```
     *
     * @category Subscription & Loading
     */
    static subscribe(id, as, depth, listener) {
        return subscribeToCoValue(this, id, as, depth, listener);
    }
    /**
     * Effectful version of `CoList.subscribe()` that returns a stream of updates.
     *
     * Needs to be run inside an `AccountCtx` context.
     *
     * @category Subscription & Loading
     */
    static subscribeEf(id, depth) {
        return subscribeToCoValueEf(this, id, depth);
    }
    /**
     * Given an already loaded `CoList`, ensure that items are loaded to the specified depth.
     *
     * Works like `CoList.load()`, but you don't need to pass the ID or the account to load as again.
     *
     * @category Subscription & Loading
     */
    ensureLoaded(depth) {
        return ensureCoValueLoaded(this, depth);
    }
    /**
     * Given an already loaded `CoList`, subscribe to updates to the `CoList` and ensure that items are loaded to the specified depth.
     *
     * Works like `CoList.subscribe()`, but you don't need to pass the ID or the account to load as again.
     *
     * Returns an unsubscribe function that you should call when you no longer need updates.
     *
     * @category Subscription & Loading
     **/
    subscribe(depth, listener) {
        return subscribeToExistingCoValue(this, depth, listener);
    }
    /** @category Type Helpers */
    castAs(cl) {
        return cl.fromRaw(this._raw);
    }
}
_a = CoList;
(() => {
    _a.prototype._type = "CoList";
})();
function toRawItems(items, itemDescriptor) {
    const rawItems = itemDescriptor === "json"
        ? items
        : "encoded" in itemDescriptor
            ? items?.map((e) => encodeSync(itemDescriptor.encoded)(e))
            : isRefEncoded(itemDescriptor)
                ? items?.map((v) => v.id)
                : (() => {
                    throw new Error("Invalid element descriptor");
                })();
    return rawItems;
}
function init(list) {
    if (list[InitValues]) {
        const { init, owner } = list[InitValues];
        const raw = owner._raw.createList(toRawItems(init, list._schema[ItemsSym]));
        Object.defineProperties(list, {
            id: {
                value: raw.id,
                enumerable: false,
            },
            _raw: { value: raw, enumerable: false },
        });
        delete list[InitValues];
    }
}
const CoListProxyHandler = {
    get(target, key, receiver) {
        if (typeof key === "string" && !isNaN(+key)) {
            const itemDescriptor = target._schema[ItemsSym];
            const rawValue = target._raw.get(Number(key));
            if (itemDescriptor === "json") {
                return rawValue;
            }
            else if ("encoded" in itemDescriptor) {
                return rawValue === undefined
                    ? undefined
                    : decodeSync(itemDescriptor.encoded)(rawValue);
            }
            else if (isRefEncoded(itemDescriptor)) {
                return rawValue === undefined
                    ? undefined
                    : new Ref(rawValue, target._loadedAs, itemDescriptor).accessFrom(receiver, Number(key));
            }
        }
        else if (key === "length") {
            return target._raw.entries().length;
        }
        else {
            return Reflect.get(target, key, receiver);
        }
    },
    set(target, key, value, receiver) {
        var _b;
        if (key === ItemsSym &&
            typeof value === "object" &&
            SchemaInit in value) {
            (_b = target.constructor)._schema || (_b._schema = {});
            target.constructor._schema[ItemsSym] =
                value[SchemaInit];
            init(target);
            return true;
        }
        if (typeof key === "string" && !isNaN(+key)) {
            const itemDescriptor = target._schema[ItemsSym];
            let rawValue;
            if (itemDescriptor === "json") {
                rawValue = value;
            }
            else if ("encoded" in itemDescriptor) {
                rawValue = encodeSync(itemDescriptor.encoded)(value);
            }
            else if (isRefEncoded(itemDescriptor)) {
                rawValue = value.id;
            }
            target._raw.replace(Number(key), rawValue);
            return true;
        }
        else {
            return Reflect.set(target, key, value, receiver);
        }
    },
    defineProperty(target, key, descriptor) {
        var _b;
        if (descriptor.value &&
            key === ItemsSym &&
            typeof descriptor.value === "object" &&
            SchemaInit in descriptor.value) {
            (_b = target.constructor)._schema || (_b._schema = {});
            target.constructor._schema[ItemsSym] =
                descriptor.value[SchemaInit];
            init(target);
            return true;
        }
        else {
            return Reflect.defineProperty(target, key, descriptor);
        }
    },
    has(target, key) {
        if (typeof key === "string" && !isNaN(+key)) {
            return Number(key) < target._raw.entries().length;
        }
        else {
            return Reflect.has(target, key);
        }
    },
};
//# sourceMappingURL=coList.js.map